
import appdirs
import datetime
import logging
import os
import selenium
import pickle


LOGGER = logging.getLogger('wabot')

USER_AGENT = "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; Trident/5.0)"
REFUSE_AFTER_EXCEPTION = True
EXECUTOR_PORT = 4444
REMOTE_EXECUTOR = 'http://127.0.0.1:%s/wd/hub'


class PickledBrowserReference:
    """
    Structure for saving a webdriver instance. Also includes the timestamp
    to help invalidate old references.
    """
    def __init__(self):
        self.browser_ref = None
        self.timestamp = None

# Factory Creator - instantiate Selenium1 and Selenium2 webdriver instances.
class CreateBrowser:
    """
    Creates and instantiates selenium webbrowser instances.

    Two strategies: remote via selenium server or local.

    The advantage of using remote is the browser instance can stay
    open after the process closes. This allows you to disconnect
    and re-connect to the same browser instance with different
    processes.
    """
    def __init__(self, pickle_filename=None):
        self._pickle_filename = None
        if pickle_filename is not None:
            self.pickle_filename = pickle_filename
        self._has_cleaned = False

    @property
    def pickle_filename(self):
        """
        Reference to the user-data-dir pickle target filename, for
        saving browser references.
        """
        if self._pickle_filename is None:
            self._pickle_filename = os.path.join(
                appdirs.user_data_dir('wabot'),
                'saved_browser_instances.pickle'
            )
            return self._pickle_filename
        else:
            return self._pickle_filename

    @pickle_filename.setter
    def pickle_filename(self, value):
        self._pickle_filename = value

    def _clean_old_pickles(self):
        if self._has_cleaned is True:
            LOGGER.debug('not cleaning pickles twice')
            return

        p = self.pickle_filename
        try:
            fp = open(p, 'rb')
            pickles = pickle.load(fp)
            if not pickles:
                raise Exception


            LOGGER.info('cleaning any old pickle refs')


            for pickle_ref in pickles:
                driver_ref = pickles[pickle_ref]
                ts = driver_ref.timestamp

                if datetime.datetime.now() - driver_ref.timestamp >= datetime.timedelta(days=3):
                    LOGGER.info('deleting old pickled driver ref: %s at %s', pickle_ref, ts)
                    pickles.pop(final_name)

            self._has_cleaned = True

        except (FileNotFoundError, IOError) as ex:
            LOGGER.error('unable to load pickles: no pickled drivers found')
        except Exception as ex:
            print(ex)
            pass

    def _load_from_pickle(self, final_name):
        p = self.pickle_filename


        # Definitely no browser instance already, we must instantiate
        if not os.path.exists(p):
            LOGGER.debug('no pickled file for saved browser instances (nothing saved yet)')
            return

        # There MAY be an open browser or an invalidated reference to a once-open browser
        if os.path.exists(p):
            LOGGER.debug('found pickled file for saved browser instances: %s', p)
            # First, see if existing session_name browser instance exists
            fp = None

            self._clean_old_pickles()
            try:
                fp = open(p, 'rb')
                # drivers = pickle.load(fp)
                pickles = pickle.load(fp)
                if not pickles:
                    raise Exception

                for idx, saved_instance in enumerate(pickles):
                    LOGGER.debug('found saved browser instances (%s): %s', idx, saved_instance)

                driver_ref = pickles.get(final_name)
                if not driver_ref:
                    raise Exception

                # if datetime.datetime.now() - driver_ref.timestamp >= datetime.timedelta(days=3):
                #     pickles.pop(final_name)

                driver = driver_ref.browser_ref
                LOGGER.debug('connected to pickled webdriver instance: %s', final_name)
                url = driver.current_url # throw error if driver isn't reliable anymore
                LOGGER.info('webdriver instance is ready')
                # self.driver = driver
                return driver
            except (FileNotFoundError, IOError) as ex:
                LOGGER.error('unable to connect to existing webdriver: no pickled drivers found')
            except Exception as ex:
                print(ex)
                self.driver = None

    def _save_to_pickle(self, final_name, driver_ref):
        p = self.pickle_filename
        pickles = {}

        # Definitely no browser instance already, we must instantiate
        if not os.path.exists(p):
            LOGGER.debug('no pickled file for saved browser instances (nothing saved yet)')

            # Create usr-app-dir if doesn't exist
            dirname = os.path.dirname(p)
            if not os.path.isdir(dirname):
                os.mkdir(dirname)

        # There MAY be an open browser or an invalidated reference to a once-open browser
        if os.path.exists(p):
            LOGGER.debug('found existing pickle file for saved browser instances: %s', p)
            # First, see if existing session_name browser instance exists


            self._clean_old_pickles()

            fp = open(p, 'rb')
            pickles = pickle.load(fp)
            if not pickles:
                raise Exception
            LOGGER.debug('found saved browser instances while saving: %s', list(pickles.keys()))
            fp.close()


        # Save to pickle - creating file if necessary
        pkl = PickledBrowserReference()
        pkl.browser_ref = driver_ref
        pkl.timestamp = datetime.datetime.now()
        pickles[final_name] = pkl

        

        import pprint;
        print('pickles:')
        pprint.pprint(pickles)

        # pickle file must be created
        fp = open(p, 'wb')
        # drivers[final_name] = driver
        LOGGER.info('saving browser instance to pickle: %s', final_name)
        pickle.dump(pickles, fp)
        fp.close()

    def _create_driver_remote_chromium(self, session_name):
        """
        Creates and returns Selenium1 chromium instance.
        """
        p = self.pickle_filename
        # e.g.: remote-chromium-webdriver
        final_name = '{}-{}'.format('remote-chromium', session_name)
        driver = None

        driver = self._load_from_pickle(final_name)
        if driver is not None:
            LOGGER.info('webdriver instance is ready (from pickle)')
            return driver

        # At this point, need to instantiate a new browser instance
        sel_host = REMOTE_EXECUTOR % (EXECUTOR_PORT)
        LOGGER.info('instantianting new browser instance (remote_chromium)')
        LOGGER.info('remote selenium: %s', sel_host)

        opt = selenium.webdriver.chrome.options.Options()
        opt.add_argument("--user-agent=" + USER_AGENT)
        opt.add_argument("--kiosk-printing")
        opt.add_argument("--focus-existing-tab-on-open=false")
        driver = selenium.webdriver.Remote(
            command_executor=sel_host,
            desired_capabilities = opt.to_capabilities()
        )

        self._save_to_pickle(final_name, driver)

        # self.driver = driver
        return driver


    # Pickle impl. is duped here

    def _create_driver_remote_firefox(self, session_name):
        """
        Creates and returns Selenium1 firefox instance.
        """
        final_name = '{}-{}'.format('remote-firefox', session_name)
        driver = None

        driver = self._load_from_pickle(final_name)
        if driver is not None:
            LOGGER.info('webdriver instance is ready (from pickle)')
            return driver

        # At this point, need to instantiate a new browser instance
        sel_host = REMOTE_EXECUTOR % (EXECUTOR_PORT)
        LOGGER.info('instantianting new browser instance (remote_firefox)')
        LOGGER.info('remote selenium: %s', sel_host)

        profile = selenium.webdriver.FirefoxProfile()
        profile.set_preference("general.useragent.override", USER_AGENT)
        driver = selenium.webdriver.Remote(
            # SELENIUM1_SERVER_PATH,
            sel_host,
            selenium.webdriver.DesiredCapabilities.FIREFOX.copy(),
            browser_profile = profile
        )

        self._save_to_pickle(final_name, driver)
        return driver



    def _create_driver_chromium2(self):
        opt = selenium.webdriver.chrome.options.Options()
        opt.add_argument("--user-agent=" + USER_AGENT)
        opt.add_argument("--kiosk-printing")
        driver = selenium.webdriver.Chrome(chrome_options = opt)
        self.driver = driver
        return driver

    def _create_driver_chromium1(self):
        # Selenium 1 - Chrome without working user agent switch
        # These two methods of creation ChromeOptions are equivalent objects
        options = selenium.webdriver.ChromeOptions()
        options.add_argument("--user-agent=" + USER_AGENT)
        driver = selenium.webdriver.Remote(desired_capabilities = options.to_capabilities())
        driver = selenium.webdriver.Remote(SELENIUM1_SERVER_PATH,
            selenium.webdriver.DesiredCapabilities.CHROME.copy())
        return driver

    def _create_driver_firefox2(self):
        # tmp = selenium.webdriver.FirefoxProfile()
        # tmp = None
        profile = None
        # filename = "/tmp/firefox_profile"
        # try:
            # fp = open(filename, "rb")
            # profile = pickle.load(fp)
        # except:
            # pass

        if not profile:
            profile = selenium.webdriver.FirefoxProfile(profile_directory = "/home/mathew/firefox_prof")
            profile.set_preference("general.useragent.override", USER_AGENT )
            profile.set_preference("browser.helperApps.neverAsk.saveToDisk", "text/csv");
            profile.set_preference("network.http.redirection-limit", "0" )
            # profile.set_preference("javascript.enabled", False )
            # profile.set_preference("print.always_print_silent", True)
            profile.set_preference("print.print_to_file", True)
            profile.set_preference("print.print_to_filename", "/tmp/print.pdf")
            profile.update_preferences()
            profile.set_preference("network.http.redirection-limit", "0" )
            # with open("/tmp/firefox_profile", "wb") as fp:
                # pickle.dump(profile, fp, pickle.HIGHEST_PROTOCOL)

        # driver = selenium.webdriver.Firefox()
        driver = selenium.webdriver.Firefox(profile)
        return driver



    def _create_driver_phantomjs(self):
        # Note(MG): Selenium support for PhantomJS has been deprecated, please use headless
        # driver = selenium.webdriver.PhantomJS()
        # return driver
        opt = selenium.webdriver.chrome.options.Options()
        opt.add_argument("--user-agent=" + USER_AGENT)
        opt.add_argument("--kiosk-printing")
        opt.add_argument("--headless")
        driver = selenium.webdriver.Chrome(chrome_options = opt)
        driver.set_window_size(838, 907)
        self.driver = driver
        return driver
